from flask import jsonify, request, current_app
from datetime import date
from ..policy import Policy
from ..models import Activity, ActivityType
from ..utils import get_policy_class, get_all_stages, get_activity_class
from . import bp
#from fasifu.GlobalConstants import GlobalConstants
#from logging import getLogger

# Flask app has its logger: current_app.logger
#logger = getLogger(GlobalConstants.loggerName)


@bp.route('/<string:lang>/<string:stage>/policy/<string:policy_number>/<string:effective_date>')
@bp.route('/<string:lang>/<string:stage>/policy/<string:policy_number>')
def get_policy(lang, stage, policy_number, effective_date=None):
    #
    # fetches a Policy by policy_number & effective_data
    # and returns it
    #

    # set default effective_date if needed
    if effective_date is None:
        effective_date = str(date.today())

    try:
        # get Policy
        policy = get_policy_class()(policy_number, effective_date)
        if policy.fetch():
            current_app.config['POLICIES'][policy.uuid] = policy
            return jsonify(policy.get()), 200
    except Exception as e:
        current_app.logger.warning(f'Fetch plolicy {policy_number} {effective_date} failed: {e}')
        return jsonify({'error': str(e)}), 400

    return jsonify({'error': 'Policy not found'}), 404

@bp.route('/stage')
def get_stages():
    #
    # returns list of all available stages
    #

    try:
        # get all stages
        stages = get_all_stages()()
        current_app.logger.debug(f"Value of stages: {stages}")
    except Exception as e:
        current_app.logger.warning(f'Failed to get All Stages: {e}')
        stages = []

    return jsonify(stages), 200


@bp.route(f'/<string:lang>/<string:stage>/activity', methods=['POST'])
def new_activity(lang, stage):
    #
    # create new activity
    #

    # get post data
    data = request.get_json()
    print(data)
    print(current_app.config.get('POLICIES'))

    # get policy and create activity 
    try:
        # get policy from app store
        policy = current_app.config['POLICIES'].get(data['id'])
        if policy is None:
            raise Exception(f'Policy {data["id"]} is absent in PoLZy storage')

        # save activity to DB
        activity = Activity.new(data, policy)

        # execute activity
        if policy.executeActivity(data['activity']):

            # update activity
            activity.finish()
            # update policy
            policy.fetch()
            if data['activity'] == "Detailauskunft":
                print(f"returned link: {policy.returnLink()}")
                return jsonify({"link": policy.returnLink()}), 200
            else:
                print(f"Activity was: {data['activity']}")

            return jsonify(policy.get()), 200
        
    except Exception as e:
        current_app.logger.warning(f'Execution activity {data.get("name")} for policy {policy.policy_number} faild: {e}')
        return jsonify({'error': 'Bad Request'}), 400

    return jsonify({
        'id': str(activity.id),
        'status': 'accepted',
        'msg': 'Activity accepted',
    }), 202



from flask import Response
from datetime import date
from . import bp
from .. import messenger

@bp.route('/listen', methods=['GET'])
def listen():

    def stream():
        messages = messenger.listen()  # returns a queue.Queue
        while True:
            msg = messages.get()  # blocks until a new message arrives
            yield msg

    return Response(
    	stream(),
    	headers={
    		"Access-Control-Allow-Origin": "*",
    		"Access-Control-Allow-Headers": "Origin, X-Requested-With, Content-Type, Accept",
    	},
    	mimetype='text/event-stream',
    )


from . import db, auth
from .auth import get_uuid, generate_token, get_expired
from datetime import datetime, timedelta, date
import uuid
import json


# authentication
@auth.verify_token
def verify_token(token):
    user = User.query.filter_by(access_key=token).first()
    if user.key_expired < datetime.utcnow():
        return user


#
# Authentication Models
#
class User(db.Model):
    __tablename__ = 'users'
    id = db.Column(db.LargeBinary, primary_key=True, default=get_uuid)
    email = db.Column(db.String(64), unique=True, nullable=False)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    displayed_name = db.Column(db.String(64), nullable=False, default='none')
    first_name = db.Column(db.String(64), nullable=True)
    last_name = db.Column(db.String(64), nullable=True)
    oauth_provider_id = db.Column(db.Integer, db.ForeignKey('oauth_providers.id'), nullable=False)
    oauth_user_id = db.Column(db.String(128), nullable=False)
    oauth_token = db.Column(db.String(128), nullable=False)
    access_key = db.Column(db.String(128), nullable=False, default=generate_token)
    key_expired = db.Column(db.DateTime, nullable=False, default=get_expired)

    # relationships
    oauth_provider = db.relationship(
        'OAuthProvider',
        foreign_keys=[oauth_provider_id],
    )

    def __str__(self):
        return self.displayed_name

    def get_name(self):
        return self.email.split('@')[0]

class OAuthProvider(db.Model):
    __tablename__ = 'oauth_providers'
    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(64), unique=True, nullable=False)
    client_id = db.Column(db.String(128), nullable=False)
    secret_key = db.Column(db.String(128), nullable=False)

    def __str__(self):
        return self.name


#
# Activity Models
#
class Activity(db.Model):
    __tablename__ = 'activities'
    id = db.Column(db.LargeBinary, primary_key=True, default=get_uuid)
    creator_id = db.Column(db.LargeBinary, db.ForeignKey('users.id'), nullable=False)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    policy_number = db.Column(db.String(64), nullable=False)
    effective_date = db.Column(db.Date, nullable=False, default=date.today)
    #type_id = db.Column(db.Integer, db.ForeignKey('activity_types.id'), nullable=False)
    type = db.Column(db.String(32), nullable=False)
    is_finished = db.Column(db.Boolean, nullable=False, default=False)
    attributes = db.Column(db.String, nullable=True)
    
    # relationships
    creator = db.relationship(
        'User',
        backref=db.backref('created_activities', order_by='desc(Activity.created)'),
        foreign_keys=[creator_id],
    )
    #type = db.relationship(
    #    'ActivityType',
    #    foreign_keys=[type_id],
    #)

    def __str__(self):
        return str(uuid.UUID(bytes=self.id))

    @classmethod
    def new(cls, data, policy):
        # 
        # create instance using data
        #

        ####### TO DELETE:
        current_user = User.query.first()
        
        # get activity type
        #activity_type = ActivityType.query.filter_by(name=data.get('activity_type')).first()
        # create activity instance
        instance = cls(
            policy_number=policy.number,
            effective_date=datetime.strptime(policy.effective_date, '%Y-%m-%d').date(),
            type=data['activity'].get('name'),
            creator=current_user,
            attributes=json.dumps(data['activity'].get('fields'))
        )
        # set effective date
        #if data.get('effective_date'):
        #    instance.effective_date = datetime.strptime(data.get('effective_date'), '%Y-%m-%d').date()
        
        # store to db
        db.session.add(instance)
        db.session.commit()
        
        return instance

    def finish(self):
        #
        # sets is_finished to True 
        #
        self.is_finished = True
        db.session.commit()

class ActivityType(db.Model):
    __tablename__ = 'activity_types'
    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(64), nullable=False)
    description = db.Column(db.String(128), nullable=True)





from datetime import datetime
import uuid


class Policy:
    #
    # Policy details, activities, attributes 
    #

    def __init__(self, number, date):
        #
        # initialization
        #

        self.id = uuid.uuid4()
        self.number = number
        self.effective_date = date
        self.data = None

    @property
    def uuid(self):
        return str(self.id)

    def fetch(self):
        #
        # IMPORTANT: this method should be define within custon implementation
        #
        # fetches policy details from Policy Management System
        #

        raise Exception('Fetch method is not set in policy class')

    def executeActivity(self, data):
        #
        # IMPORTANT: this method should be define within custon implementation
        #
        # executes policy activity defined in data
        #

        raise Exception('Execute Activity method is not set in policy class')

    def get(self):
        #
        # returns Policy data
        #

        result = {'id': self.uuid}
        if self.data:
            result.update(self.data)

        return result

    
    #def get_status(self):
    #    #
    #    # returns status of policy
    #    #
    #    if self.data:
    #        return self.data.get('status')

    #def get_product_line(self):
    #    #
    #    # returns name of policy's product line
    #    #
    #
    #    if self.data and self.data.get('product_line'):
    #        return self.data.prosuct_line.get('name')

    #def get_insured_is_person(self):
    #    #
    #    # returns:
    #    # - True if insured person
    #    # - False if insured object
    #    # - None otherwise
    #    #
    #
    #    if self.data and self.data.get('insured_object'):
    #        return self.data['insured_object'].get('is_person')

    #def get_insured_object_type(self):
    #    #
    #    # returns type of insure object
    #    #
    #
    #    if self.get_insured_is_person() is False:
    #        return self.data['insured_object'].get('type')


    #def get_possible_activities(self):
    #    #
    #    # returns list of possible activitites for policy
    #    #
    #
    #    return self.activities_by_state.get(self.get_status())

    #def get_attribute_descriptions(self):
    #    #
    #    # returns name-description of attributes applicable to policy 
    #    #
    #
    #    return {
    #        'policy': self.attributes_policy,
    #        'product_line': self.attributes_product_line.get(self.data['product_line'].get('name')),
    #        'insured_object': self.attributes_insured_person if self.get_insured_is_person() else self.attributes_insured_object,
    #        'insured_object_type': self.attributes_insured_object_type.get(self.get_insured_object_type()),
    #        'implementation': self.attributes_implementation,
    #    }

    



import queue

class Messenger:

    def __init__(self):
        self.listeners = []

    def listen(self):
        q = queue.Queue(maxsize=5)
        self.listeners.append(q)
        return q

    def announce(self, msg):
        for i in reversed(range(len(self.listeners))):
            try:
                self.listeners[i].put_nowait(msg)
            except queue.Full:
                del self.listeners[i]

    @classmethod
    def format_sse(cls, data: str, event=None) -> str:
        msg = f'data: {data}\n\n'
        if event is not None:
            msg = f'event: {event}\n{msg}'
        return msg


#
# populates PoLZy db with sample instances
#

from . import db
from .models import User, ActivityType, OAuthProvider
from .auth import generate_token
from .policy import Policy
import uuid
from datetime import datetime, timedelta


# create OAuth provider
print('Creating OAuth Provider...')
provider = OAuthProvider(
    name='SampleProvider',
    client_id=str(uuid.uuid4()),
    secret_key=generate_token(16),
)
db.session.add(provider)

# create users
print('Creating users...')
user = User(
    email='admin@sample.com',
    oauth_provider=provider,
    oauth_user_id=str(uuid.uuid4()),
    oauth_token=generate_token(16),
    key_expired=datetime.now() + timedelta(days=360),
)
db.session.add(user)

# create activity types
print('Creating Activity Types...')
activities = {}
for name in {i for key, value in Policy.activities_by_status.items() for i in value}:
    activity = ActivityType(
        name=name,
        description=f'Sample {name.capitalize()} Activity',
    )
    db.session.add(activity)

# save to db
db.session.commit()
print('Done.')



from flask import Flask
from flask_sqlalchemy import SQLAlchemy
from flask_migrate import Migrate
from flask_httpauth import HTTPTokenAuth
import os
from .messenger import Messenger


# initialization
db = SQLAlchemy()
migrate = Migrate()
auth = HTTPTokenAuth(scheme='Bearer')
messenger = Messenger()


def create_app(config=None):
    # create application
    app = Flask(__name__)
    # set default config
    app.config['SECRET_KEY'] = os.getenv('SECRET_KEY', default = 'secret!key')
    app.config['JSON_SORT_KEYS'] = False
    app.config['SQLALCHEMY_DATABASE_URI'] = os.getenv(
        'DATABASE_URL',
        default='sqlite:///' + os.path.join(os.path.abspath(os.path.dirname(__file__)), 'polzy.db'),
    )
    app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False
    app.config.from_object(config)

    # policy store
    app.config['POLICIES'] = {}

    db.init_app(app)
    migrate.init_app(app, db)

    from .main import bp
    app.register_blueprint(bp)

    from .announce import bp as bp_announce
    app.register_blueprint(bp_announce)

    return app



from . import models
